@@ -1371,14 +1371,14 @@ static void LZe_full_flush(CLzmaEnc *p, uint32_t posState)
     {
     unsigned long long in_size = p->nowPos64;
     unsigned long long out_size = p->rc.processed + Fh_size + Ft_size;
-    if( in_size <= 0 || out_size <= 0 )
+    if( in_size == 0 || out_size == 0 )
       fputs( " no data compressed.\n", stderr );
     else
-      fprintf( stderr, "%6.3f:1, %6.3f bits/byte, "
-                       "%5.2f%% saved, %llu in, %llu out.\n",
+      fprintf( stderr, "%6.3f:1, %5.2f%% ratio, %5.2f%% saved, "
+                       "%llu in, %llu out.\n",
                (double)in_size / out_size,
-               ( 8.0 * out_size ) / in_size,
-               100.0 * ( 1.0 - ( (double)out_size / in_size ) ),
+               ( 100.0 * out_size ) / in_size,
+               100.0 - ( ( 100.0 * out_size ) / in_size ),
                in_size, out_size );
     }
   }
